
__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Exception classes
"""

import sys
import weakref
import operator
import traceback


# We have the choice of doing signature checks in exceptions, or to raise saga
# exceptions on signature checks -- we cannot do both.  At this point, we use
# the saga.exceptions in signatures, thus can *not* have signature checks
# here...

# ------------------------------------------------------------------------------
#
class SagaException (Exception) :
    """
    The Exception class encapsulates information about error conditions
    encountered in SAGA.

    Additionally to the error message (e.message), the exception also provides
    a trace to the code location where the error condition got raised
    (e.traceback).

    B{Example}::

      try :
          file = saga.filesystem.File ("sftp://comet.xsede.org/tmp/data1.dat")

      except saga.Timeout as to :
          # maybe the network is down?
          print("connection timed out")
      except saga.Exception as e :
          # something else went wrong
          print("Exception occurred: %s %s" % (e, e.traceback))

    There are cases where multiple backends can report errors at the same time.
    In that case, the radical.saga implementation will collect the exceptions,
    sort them by their 'rank', and return the highest ranked one.  All other
    catched exceptions are available via :func:`get_all_exceptions`, or via the
    `exceptions` property.

    The rank of an exception defines its explicity: in general terms: the
    higher the rank, the better defined / known is the cause of the problem.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        """
        Create a new exception object.

        :param msg:         The exception message.
        :param parent:      Original exception
        :param api_object:  The object that has caused the exception, default is
                            None.
        :param from_log:    Exception c'tor originates from the static log_
                            member method (ignore in exception stack!)
        """
        Exception.__init__(self, msg)

        self._plain_message = msg
        self._exceptions    = [self]
        self._top_exception = self
        self._ptype         = type(parent).__name__   # parent exception type
        self._stype         = type(self  ).__name__   # own exception    type

        ignore_stack = 2
        if  from_log :
            ignore_stack += 1


        if api_object :
            self._object    = weakref.ref (api_object)
        else :
            self._object    = None


        # did we get a parent exception?
        if  parent :

            # if so, then this exception is likely created in some 'except'
            # clause, as a reaction on a previously catched exception (the
            # parent).  Thus we append the message of the parent to our own
            # message, but keep the parent's traceback (after all, the original
            # exception location is what we are interested in).
            #
            if  isinstance (parent, SagaException) :
                # that all works nicely when parent is our own exception type...
                self._traceback = parent.traceback

                frame           = traceback.extract_stack ()[- ignore_stack]
                line            = "%s +%s (%s)  :  %s" % (frame.filename,
                    frame.lineno, frame.name, frame.line)
                self._message   = "  %-20s: %s (%s)\n%s" \
                                % (self._stype, msg, line, parent.msg)

            else:
                if self._stype != "NoneType" :
                    # ... but if parent is a native (or any other) exception
                    # type, we don't have a traceback really -- so we dig it
                    # out of sys.exc_info.
                    trace           = sys.exc_info ()[2]
                    stack           = traceback.extract_tb  (trace)
                    traceback_list  = traceback.format_list (stack)
                    self._traceback = "".join (traceback_list)

                    # the message composition is very similar -- we just inject
                    # the parent exception type inconspicuously somewhere (above
                    # that was part of 'parent.message' already).
                    frame          = traceback.extract_stack ()[- ignore_stack]
                    line           = "%s +%s (%s)  :  %s" % (frame.filename,
                        frame.lineno, frame.name, frame.line)
                    self._message  = "  %-20s: %s (%s)\n  %-20s: %s" \
                                % (self._stype, msg, line, self._ptype, parent)

        else :

            # if we don't have a parent, we are a 1st principle exception,
            # i.e. a reaction to some genuine code error.  Thus we extract the
            # traceback from exactly where we are in the code (the last stack
            # frame will be the call to this exception constructor), and we
            # create the original exception message from 'stype' and 'message'.
            stack           = traceback.extract_stack ()
            traceback_list  = traceback.format_list (stack)
            self._traceback = "".join (traceback_list[:-1])
            frame           = traceback.extract_stack ()[- ignore_stack -1]
            line            = "%s +%s (%s)  :  %s" % (frame.filename,
                frame.lineno, frame.name, frame.line)
            self._message   = "%s (%s)" % (msg, line)

        # we can't do that earlier as _msg was not set up before
        self._messages = [self._message]


    # --------------------------------------------------------------------------
    #
    def __str__ (self) :
        return self.get_message ()


    # --------------------------------------------------------------------------
    #
    def __repr__ (self) :
        return "%s\n%s" % (self._message, self._traceback)


    # --------------------------------------------------------------------------
    #
    def _clone (self) :
        """ method is used internally -- see :func:`_get_exception_stack`."""

        clone = self.__class__ ("")

        clone._parent    = self._parent
        clone._object    = self._object
        clone._message   = self._message
        clone._messages  = self._messages
        clone._exception = self._exceptions
        clone._traceback = self._traceback
        clone._stype     = self._stype
        clone._ptype     = self._ptype

        return clone

    # --------------------------------------------------------------------------
    #
    @classmethod
    def _log (cls, logger, msg, parent=None, api_object=None, level='error'):
        """
        This class method allows to log the exception message while
        constructing a SAGA exception, like::

          # raise an exception, no logging
          raise saga.IncorrectState("File closed")

          # raise an exception, log as error event (error level is default)
          raise saga.IncorrectState._log(logger, "File closed")

          # raise an exception, log as warning event
          raise saga.IncorrectState._log(logger, "File closed", level=warning)
          raise saga.IncorrectState._log(logger, "File closed", warning) # same

        This way, the 'raise' remains clearly in the code, as that is the
        dominating semantics of the call.
        """

        log_method = logger.error

        try :
            log_method = getattr (logger, level.lower())
        except :
            sys.stderr.write ("unknown log level '%s'"  %  level)

        log_method ("%s: %s" % (cls.__name__, msg))

        return cls (msg, parent=parent, api_object=api_object, from_log=True)



    # --------------------------------------------------------------------------
    #
    def get_message (self) :
        """ Return the exception message as a string.  That message is also
        available via the 'message' property."""
        return self._message


    # --------------------------------------------------------------------------
    #
    def _get_plain_message (self) :
        """ Return the plain error message as a string. """
        return self._message


    # --------------------------------------------------------------------------
    #
    def get_type (self):
        """ Return the type of the exception as string.
        """
        return self._stype


    # --------------------------------------------------------------------------
    #
    def get_object (self) :
        """ Return the object that raised this exception. An object may not
        always be available -- for example, exceptions raised during object
        creation may not have the option to keep an incomplete object instance
        around.  In those cases, this method will return 'None'.  Either way,
        the object is also accessible via the 'object' property.
        """

        # object is a weak_ref, and may have been garbage collected - we simply
        # return 'None' then
        return self._object ()


    # --------------------------------------------------------------------------
    #
    def _add_exception (self, e) :
        """
        Some sub-operation raised a SAGA exception, but other exceptions may be
        catched later on.  In that case the later exceptions can be added to the
        original one with :func:`_add_exception`\(e).  Once all exceptions are
        collected, a call to :func:`_get_exception_stack`\() will return a new
        exception which is selected from the stack by rank and order.  All other
        exceptions can be accessed from the returned exception by
        :func:`get_all_exceptions`\() -- those exceptions are then also ordered
        by rank.
        """

        self._exceptions.append (e)
        self._messages.append   (e.message)

        if e._rank > self._top_exception._rank :
            self._top_exception = e


    # --------------------------------------------------------------------------
    #
    def _get_exception_stack (self) :
        """
        This method is internally used by the radical.saga engine, and is only
        relevant for operations which (potentially) bind to more than one
        adaptor.
        """

        if self._top_exception == self :
            return self

        # damned, we can't simply recast ourself to the top exception type -- so
        # we have to create a new exception with that type, and copy all state
        # over...

        # create a new exception with same type as top_exception
        clone = self._top_exception._clone ()
        clone._exceptions = []
        clone._messages   = []

        # copy all state over
        for e in sorted(self._exceptions, key=operator.attrgetter ('_rank'),
                                          reverse=True):
            clone._exceptions.append (e)
            clone._messages.append (e._message)

        return clone


    # --------------------------------------------------------------------------
    #
    def get_all_exceptions (self) :
        return self._exceptions


    # --------------------------------------------------------------------------
    #
    def get_all_messages (self) :
        return self._messages


    # --------------------------------------------------------------------------
    #
    def get_traceback (self) :
        return self._traceback


    # --------------------------------------------------------------------------
    #
    message        = property (get_message)         # string
    object         = property (get_object)          # object type
    type           = property (get_type)            # exception type
    exceptions     = property (get_all_exceptions)  # list [Exception]
    messages       = property (get_all_messages)    # list [string]
    traceback      = property (get_traceback)       # string


# ------------------------------------------------------------------------------
#
class NotImplemented(SagaException, NotImplementedError):
    """ radical.saga does not implement this method or class. (rank: 11)"""

    _rank = 11

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class IncorrectURL(SagaException, ValueError):
    """ The given URL could not be interpreted, for example due to an incorrect
        / unknown schema. (rank: 10)"""

    _rank = 10

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class BadParameter(SagaException, ValueError):
    """ A given parameter is out of bound or ill formatted. (rank: 9)"""

    _rank = 9

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class AlreadyExists(SagaException, ValueError):
    """ The entity to be created already exists. (rank: 8)"""

    _rank = 8

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class DoesNotExist(SagaException, KeyError):
    """ An operation tried to access a non-existing entity. (rank: 7)"""

    _rank = 7

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class IncorrectState(SagaException, RuntimeError):
    """ The operation is not allowed on the entity
        in its current state. (rank: 6)"""

    _rank = 6

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class PermissionDenied(SagaException, RuntimeError):
    """ The used identity is not permitted to perform the
        requested operation. (rank: 5)"""

    _rank = 5

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class AuthorizationFailed(SagaException, RuntimeError):
    """ The backend could not establish a valid identity. (rank: 4)"""

    _rank = 4

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class AuthenticationFailed(SagaException, RuntimeError):
    """ The backend could not establish a valid identity. (rank: 3)"""

    _rank = 3

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class Timeout(SagaException, RuntimeError):
    """ The interaction with the backend times out. (rank: 2)"""

    _rank = 2

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------
#
class NoSuccess(SagaException, RuntimeError):
    """ Some other error occurred. (rank: 1)"""

    _rank = 1

    def __init__ (self, msg, parent=None, api_object=None, from_log=False) :
        SagaException.__init__ (self, msg, parent, api_object, from_log)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"

import traceback
import radical.utils as ru

from ..import exceptions as se


# ------------------------------------------------------------------------------
#
def translate_exception (e, msg=None) :
    """
    In many cases, we should be able to roughly infer the exception cause
    from the error message -- this is centrally done in this method.  If
    possible, it will return a new exception with a more concise error
    message and appropriate exception type.
    """

    if  not issubclass (e.__class__, se.SagaException) :
        # we do not touch non-saga exceptions
        return e

    if  not issubclass (e.__class__, se.NoSuccess) :
        # this seems to have a specific cause already, leave it alone
        return e

    ru.Logger('radical.saga.pty').debug (traceback.format_exc())


    cmsg = e._plain_message

    if  msg :
        cmsg = "%s (%s)" % (cmsg, msg)

    lmsg = cmsg.lower ()

    if   'could not resolve hostname' in lmsg: e = se.BadParameter (cmsg)
    elif 'connection timed out'       in lmsg: e = se.BadParameter (cmsg)
    elif 'connection refused'         in lmsg: e = se.BadParameter (cmsg)
    elif 'auth'                       in lmsg: e = se.AuthorizationFailed(cmsg)
    elif 'pass'                       in lmsg: e = se.AuthenticationFailed(cmsg)
    elif 'denied'                     in lmsg: e = se.PermissionDenied (cmsg)
    elif 'man-in-the-middle'          in lmsg: 
        e = se.AuthenticationFailed("ssh key mismatch: %s" % cmsg)
    elif 'ssh_exchange_identification' in lmsg :
        e = se.AuthenticationFailed ("too many login attempts: %s" % cmsg)
    elif 'shared connection' in lmsg :
        e = se.NoSuccess ("Insufficient system resources: %s" % cmsg)
    elif 'pty allocation' in lmsg :
        e = se.NoSuccess ("Insufficient system resources: %s" % cmsg)
    elif 'Connection to master closed' in lmsg :
        e = se.NoSuccess ("Connection closed by system: %s" % cmsg)

    return e


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils               as ru
import radical.utils.signatures    as rus

from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base       as sab

from .. import sasync
from .. import task                as st
from .. import base                as sb
from .. import session             as ss
from .. import exceptions          as se
from .. import attributes          as sa
from .. import constants           as sc

from .  import description         as descr
from .  import constants           as c


# ------------------------------------------------------------------------------
#
class Resource (sb.Base, sa.Attributes, sasync.Async) :
    """
    A :class:`Resource` class instance represents a specific slice of resource
    which is, if in `RUNNING` state, under the applications control and ready to
    serve usage requests.  The type of accepted usage requests depends on the
    specific resource types (job execution for :class:`saga.resource.Compute`,
    data storage for :class:`saga.resource.Storage`, and network connectivity
    for :class:`saga.resource.Network`.  The exact mechanism how those usage
    requests are communicated are not part of the resource's class interface,
    but are instead served by other RADICAL-SAGA classes -- typically those are
    :class:`saga.job.Service` for Compute resources, and
    :class:`saga.filesystem.Directory` for Storage resources (Network resources
    provide implicit connectivity, but do not have explicit, public entry points
    to request usage.

    The process of resource acquisition is performed by a *ResourceManager*,
    represented by a :class:`saga.resource.Manager` instance.  The semantics of
    the acquisition process is defined as the act of moving a slice (subset) of
    the resources managed by the resource manager under the control of the
    requesting application (i.e. under user control), to use as needed.  The
    type and property of the resource slice to be acquired and the time and
    duration over which the resource will be made available to the application
    are specified in a :class:`saga.resource.Description`, to be supplied when
    acquiring a resource.

    The exact backend semantics on *how* a resource slice is provisioned to the
    application is up to the resource manager backend -- this can be as simple
    as providing a job submission endpoint to a classic HPC resource, and as
    complex as instantiating a pilot job or pilot data container, or reserving
    a network fiber on demand, or instantiating a virtual machine -- the result
    will, from the application's perspective, indistinguishable: a resource
    slice is made available for the execution of usage requests (tasks,
    workload, jobs, ...).

    Resources are stateful: when acquired from a resource manager, they are
    typically in `NEW` state, and will become `ACTIVE` once they are provisioned
    to the application and can serve usage requests.  Some resources may  go
    through an intermediate state, `PENDING`, when they are about to become
    active at some point, and usage requests can already be submitted -- those
    usage requests will not be executed until the resources enters the `ACTIVE`
    state.  The resource can be release from application control in three
    different ways: they can be actively be destroyed by the application, and
    will then enter the `CANCELED` state; they can internally cease to function
    and become unable to serve usage requests, represented by a `FAILED` state,
    and the resource manager can retract control from the application because
    the agreed time duration has passed -- this is represented by the `EXPIRED`
    state.
    """
    # FIXME:
    #   - we don't use PENDING like this, yet
    #   - include state diagram (also for jobs btw)

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource',
                  rus.optional (str),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, id=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        """
        __init__(id=None, session=None)

        Create / reconnect to a resource.

        :param id: id of the resource
        :type  id: :class:`saga.Url`

        :param session: :class:`saga.Session`

        Resource class instances are usually created by calling :func:`acquire`
        on the :class:`saga.resource.Manager` class.  Already acquired resources
        are identified by a string typed identifier.  This constructor accepts
        such an identifier to create another representation of the same
        resource.  As the resource itself is new newly acquired, it can be in
        any state.  In particular, it can be in a final state, and thus be
        unusable.  Further, the resource may already have expired or failed, and
        the information about it may have been purged -- in that case the id
        will not be valid any longer, and a :class:`saga.BadParameter` exception
        will be raised.

        The session parameter is interpreted exactly as the session parameter on
        the :class:`saga.resource.Manager` constructor.
        """

        # set attribute interface properties

        import radical.saga.attributes as sa

        self._attributes_extensible  (False)
        self._attributes_camelcasing (True)

        # register properties with the attribute interface

        self._attributes_register(c.ID          , None, sa.ENUM,   sa.SCALAR, sa.READONLY)
        self._attributes_register(c.RTYPE       , None, sa.ENUM,   sa.SCALAR, sa.READONLY)
        self._attributes_register(c.STATE       , None, sa.ENUM,   sa.SCALAR, sa.READONLY)
        self._attributes_register(c.STATE_DETAIL, None, sa.STRING, sa.SCALAR, sa.READONLY)
        self._attributes_register(c.ACCESS      , None, sa.URL,    sa.SCALAR, sa.READONLY)
        self._attributes_register(c.MANAGER     , None, sa.URL,    sa.SCALAR, sa.READONLY)
        self._attributes_register(c.DESCRIPTION , None, sa.ANY,    sa.SCALAR, sa.READONLY)

        self._attributes_set_enums (c.STATE, [c.UNKNOWN ,
                                              c.PENDING ,
                                              c.ACTIVE  ,
                                              c.CANCELED,
                                              c.EXPIRED ,
                                              c.FAILED  ,
                                              c.FINAL   ])

        self._attributes_set_enums (c.RTYPE, [c.COMPUTE ,
                                              c.STORAGE ,
                                              c.NETWORK ])


        self._attributes_set_getter (c.ID          , self.get_id          )
        self._attributes_set_getter (c.RTYPE       , self.get_rtype       )
        self._attributes_set_getter (c.STATE       , self.get_state       )
        self._attributes_set_getter (c.STATE_DETAIL, self.get_state_detail)
        self._attributes_set_getter (c.ACCESS      , self.get_access      )
        self._attributes_set_getter (c.MANAGER     , self.get_manager     )
        self._attributes_set_getter (c.DESCRIPTION , self.get_description )


        # FIXME: we need the ID to be or to include an URL, as we don't have
        # a scheme otherwise, which means we can't select an adaptor.  Duh! :-/

        # FIXME: documentation for attributes is missing.

        # param checks
        scheme = None
        if  not id :
            if  'resource_schema' not in _adaptor_state :
                raise se.BadParameter ("Cannot initialize resource without id"
                                    % self.rtype)
            else :
                scheme = _adaptor_state['resource_schema']
        else :
            # ID is formatted as '[manager-url]-[resource-id]'
            import parse   # FIXME: use regex to reduce number of dependencies
            res      = parse.parse('[{}]-[{}]', id)
            url      = ru.Url(res[0])
            scheme   = url.scheme.lower ()

        if not session :
            session = ss.Session (default=True)

        self._base  = super (Resource, self)
        self._base.__init__ (scheme, _adaptor, _adaptor_state, id, session,
                             ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('resource',
                  rus.optional (str),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create   (cls, id=None, session=None, ttype=sc.SYNC) :
        """
        This is the asynchronous class constructor, returning
        a :class:`saga:Task` instance.  For details on the accepted parameters,
        please see the description of :func:`__init__`.
        """

        return cls (id, session, _ttype=ttype)._init_task


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource',
                  descr.Description,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def reconfig (self, descr, ttype=None) :
        """
        reconfig(descr)

        A resource is acquired according to a resource description, i.e. to
        a specific set of attributes.  At some point in time, while the
        resource is running, the application requirements on the resource may
        have changed -- in that case, the application can request to change the
        resource's configuration on the fly.

        This method cannot be used to change the type of the resource.  Backends
        may or may not support this operation -- if not,
        a :class:`saga.NotImplemented` exception is raised.  If the method is
        supported, , then the semantics of the method is equivalent to the
        semantics of the :func:`acquire` call on the
        :class:`saga.resource.Manager` class.
        """

        return self._adaptor.reconfig (descr, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def destroy  (self, ttype=None) :
        """
        destroy()

        The semantics of this method is equivalent to the semantics of the
        :func:`destroy` call on the :class:`saga.resource.Manager` class.
        """

        return self._adaptor.destroy (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource',
                  rus.optional (rus.one_of (c.UNKNOWN, c.NEW,      c.PENDING,
                                            c.ACTIVE,  c.DONE,     c.FAILED,
                                            c.EXPIRED, c.CANCELED, c.FINAL)),
                  rus.optional (float),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def wait (self, state=c.FINAL, timeout=None, ttype=None) :
        """
        wait(state=FINAL, timeout=None)

        Wait for a resource to enter a specific state.

        :param state: resource state to wait for (UNKNOWN, NEW, PENDING, ACTIVE,
                      DONE, FAILED, EXPIRED, CANCELED, FINAL)

        :type  state: float
        :param state: time to block while waiting.

        This method will block until the resource entered the specified state,
        or until `timeout` seconds have passed -- whichever occurs earlier.  If
        the resource is in a final state, the call will raise and
        :class:`saga.IncorrectState` exception when asked to wait for any
        non-final state.

        A negative `timeout` value represents an indefinit timeout.
        """

        # FIXME:
        #   - right now, we can not put a resource in a `TaskContainer`, because
        #     it is not a `Task`.  We need to either reconsider the class
        #     hierarchy, and then inherit a `ResourceContainer` from
        #     `TaskContainer` (see job); or we implement a `ResourceContainer`
        #     from scratch...

        return self._adaptor.wait (state, timeout, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, str, st.Task))
    def get_id   (self, ttype=None) :
        """
        get_id()

        Return the resource ID.
        """
        return self._adaptor.get_id            (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes    ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns  ((rus.one_of (c.COMPUTE,
                                c.STORAGE,
                                c.NETWORK), st.Task))
    def get_rtype (self, ttype=None) :
        """
        get_rtype()

        Return the resource type.
        """
        return self._adaptor.get_rtype         (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes    ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns  ((rus.one_of (c.UNKNOWN ,
                                c.NEW     ,
                                c.PENDING ,
                                c.ACTIVE  ,
                                c.CANCELED,
                                c.EXPIRED ,
                                c.DONE    ,
                                c.FAILED  ,
                                c.FINAL   ), st.Task))
    def get_state (self, ttype=None) :
        """
        get_state()

        Return the state of the resource.
        """
        return self._adaptor.get_state         (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, str, st.Task))
    def get_state_detail (self, ttype=None) :
        """
        get_state_detail()

        Return the state details (backend specific) of the resource.
        """
        return self._adaptor.get_state_detail  (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((rus.nothing, str, st.Task))
    def get_access (self, ttype=None) :
        """
        get_access()

        Return the resource access Url.
        """
        return self._adaptor.get_access        (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes      ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns    ((str, st.Task))
    def get_manager (self, ttype=None) :
        """
        get_manager()

        Return the manager instance that was used to acquire this resource.
        """
        return self._adaptor.get_manager       (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Resource', rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, descr.Description, st.Task))
    def get_description  (self, ttype=None) :
        """
        get_description()

        Return the description that was used to aquire this resource.
        """
        return self._adaptor.get_description   (ttype=ttype)


# ------------------------------------------------------------------------------
#
class Compute (Resource) :
    """
    A Compute resource is a resource which provides compute capabilities, i.e.
    which can execute compute jobs.  As such, the 'Access' attribute of the
    compute resource (a URL) can be used to create a :class:`saga.job.Service`
    instance to submit jobs to.
    """

    # --------------------------------------------------------------------------
    # FIXME: should 'ACCESS' be a list of URLs?  A VM could have an ssh *and*
    #        a gram endpoint...

    # @rus.takes   ('ComputeResource',
    #               rus.optional (basestring),
    #               rus.optional (ss.Session),
    #               rus.optional (sab.Base),
    #               rus.optional (dict),
    #               rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    # @rus.returns (rus.nothing)
    def __init__ (self, id=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :

        self._resrc = super  (Compute, self)
        self._resrc.__init__ (id, session, _adaptor, _adaptor_state, _ttype)

        if  self.rtype != c.COMPUTE :
            raise se.BadParameter ("Cannot init Compute resource type %s"
                                % self.rtype)


# ------------------------------------------------------------------------------
#
class Storage (Resource) :
    """
    A Storage resource is a resource which has storage capabilities, i.e. the
    ability to persistently store, organize and retrieve data.  As such, the
    'Access' attribute of the storage resource (a URL) can be used to create
    a :class:`saga.filesystem.Directory` instance to manage the resource's data
    space.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('StorageResource',
                  rus.optional (str),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, id=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :

        self._resrc = super  (Storage, self)
        self._resrc.__init__ (id, session, _adaptor, _adaptor_state, _ttype)

        if  self.rtype != c.STORAGE :
            raise se.BadParameter ("Cannot init Storage resource type %s"
                                % self.rtype)


# ------------------------------------------------------------------------------
#
class Network (Resource) :
    """
    A Network resource is a resource which has network capabilities.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('NetworkResource',
                  rus.optional (str),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, id=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :

        self._resrc = super  (Network, self)
        self._resrc.__init__ (id, session, _adaptor, _adaptor_state, _ttype)

        if  self.rtype != c.NETWORK :
            raise se.BadParameter ("Cannot init Network resource type %s"
                                % self.rtype)
#
# ------------------------------------------------------------------------------








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"

import copy

import radical.utils            as ru
import radical.utils.signatures as rus

from . import exceptions               as se

from .engine import engine
from . import context
from . import base


# ------------------------------------------------------------------------------
#
class _ContextList (list) :
    """
    The `saga.Session` class has a 'contexts' member which is a mutable list of
    `saga.Context` instances.  Adding contexts to that list is semantically
    equivalent to calling `saga.Session.add_context (ctx)`, which (a) creates
    a deep copy of the context, and (b) initializes the context (i.e. calls
    `ctx._initialize (self)`, with `self` being the session instance).  We thus
    create our own provate `_ContextList` class which inherits from the native
    Python `list` class, and overload the `append()` call with said semantics.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, session=None, *args, **kwargs) :

        self._session = session

        if  session :
            self._logger  = session._logger
        else :
            self._logger  = ru.Logger('radical.saga')

        base_list = super  (_ContextList, self)
        base_list.__init__ (*args, **kwargs)

    # --------------------------------------------------------------------------
    #
    def append (self, ctx, session=None) :

        ctx_clone = self._initialise_context(ctx, session)

        # context initialized ok, add it to the list of known contexts
        super (_ContextList, self).append (ctx_clone)

    # --------------------------------------------------------------------------
    #
    def insert(self, index, ctx, session=None) :

        ctx_clone = self._initialise_context(ctx, session)

        # context initialized ok, add it to the list of known contexts
        super (_ContextList, self).insert (0, ctx_clone)


    # --------------------------------------------------------------------------
    # Initialise a context to be added to the list of known contexts
    # Returns a cloned, initialised context that can be added to the context
    # list.
    def _initialise_context(self, ctx, session=None):

        if  not isinstance (ctx, context.Context) :
            raise TypeError("item to add is not a saga.Context instance")

        # create a deep copy of the context (this keeps _adaptor etc)
        ctx_clone = context.Context  (ctx.type)
        ctx._attributes_deep_copy (ctx_clone)

        if  not session :
            session = self._session
            logger  = self._logger
        else :
            logger  = session._logger


        # try to initialize that context, i.e. evaluate its attributes and
        # infer additional runtime information as needed
      # logger.debug ("adding  context : %s" % (ctx_clone))

        if  not session :
            logger.warning ("cannot initialize context - no session: %s" \
                       % (ctx_clone))
        else :
            try :
                ctx_clone._initialize (session)
            except se.SagaException as e:
                msg = "Cannot add context, initialization failed (%s)"  %  str(e)
                raise se.BadParameter (msg) from e

        return ctx_clone

    # --------------------------------------------------------------------------
    #
    def __deepcopy__(self, memo):

        ret = _ContextList()

        for c in self:
            ret.append(c)

        return ret


# ------------------------------------------------------------------------------
#
class Session (base.SimpleBase) :
    """A SAGA Session object as defined in GFD.90.

    A SAGA session has the purpose of scoping the use of security credentials
    for remote operations.  In other words, a session instance acts as
    a container for security Context instances -- SAGA objects (such as
    job.Service or filesystem.File) created in that session will then use
    exactly the security contexts from that session (and no others).

    That way, the session serves two purposes:  (1) it helps SAGA to decide
    which security mechanism should be used for what interaction, and (2) it
    helps SAGA to find security credentials which would be difficult to pick up
    automatically.

    The use of a session is as follows:


    Example::


        # define an ssh context
        c = saga.Context('ssh')
        c.user_cert = '$HOME/.ssh/special_id_rsa.pub'
        c.user_key  = '$HOME/.ssh/special_id_rsa'

        # add it to a session
        s = saga.Session
        s.add_context(c)

        # create a job service in this session -- that job service can now
        # *only* use that ssh context.
        j = saga.job.Service('ssh://remote.host.net/', s)


    The session argument to the L{job.Service} constructor is fully optional --
    if left out, SAGA will use default session, which picks up some default
    contexts as described above -- that will suffice for the majority of use
    cases.

    A session instance exposes a `context` property, which is a list of
    authentication contexts managed by this session.  As the contexts and the
    session are stateless, it is safe to modify this list as needed.
    """

    # FIXME: session deep copy not implemented

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Session',
                  rus.optional(bool))
    @rus.returns (rus.nothing)
    def __init__ (self, default=True, uid=None):
        """
        default: bool
        ret:     None
        """

        simple_base = super  (Session, self)
        simple_base.__init__ (uid=uid)

        self._logger = ru.Logger('radical.saga')

        # if the default session is expected, we point our context list to the
        # shared list of the default session singleton.  Otherwise, we create
        # a private list which is not populated.

        # a session also has a lease manager, for adaptors in this session to use.

        if  default :
            default_session     = DefaultSession (uid=self._id)
            self.contexts       = copy.deepcopy(default_session.contexts)
            self._lease_manager = default_session._lease_manager
        else :
            self.contexts       = _ContextList (session=self)

            # FIXME: at the moment, the lease manager is owned by the session.
            # Howevwer, the pty layer is the main user of the lease manager,
            # and we thus keep the lease manager options in the pty subsection.
            # So here we are, in the session, evaluating the pty config options.
            self._cfg = ru.Config(module='radical.saga.session')
            self._lease_manager = ru.LeaseManager (
                    max_pool_size=self._cfg.pty.connection_pool_size,
                    max_pool_wait=self._cfg.pty.connection_pool_wait,
                    max_obj_age  =self._cfg.pty.connection_pool_ttl
                    )


    # ----------------------------------------------------------------
    #
    @rus.takes   ('Session')
    @rus.returns (str)
    def __str__  (self):
        """String represenation."""

        return "Registered contexts: %s" % (str(self.contexts))


    # ----------------------------------------------------------------
    #
    @rus.takes      ('Session',
                     context.Context)
    @rus.returns    (rus.nothing)
    def add_context (self, ctx) :
        """
        ctx:     saga.Context
        ret:     None

        Add a security L{Context} to the session.
        It is encouraged to use the L{contexts} property instead.
        """

        return self.contexts.insert (0, ctx=ctx, session=self)


    # ----------------------------------------------------------------
    #
    @rus.takes   ('Session',
                  context.Context)
    @rus.returns (rus.nothing)
    def remove_context (self, ctx) :
        """
        ctx:     saga.Context
        ret:     None

        Remove a security L{Context} from the session.
        It is encouraged to use the L{contexts} property instead.
        """

        if ctx in self.contexts :
            self.contexts.remove (ctx)


    # ----------------------------------------------------------------
    #
    @rus.takes   ('Session')
    @rus.returns (rus.list_of (context.Context))
    def list_contexts  (self) :
        """
        ret:     list[saga.Context]

        Retrieve all L{Context} objects attached to the session.
        It is encouraged to use the L{contexts} property instead.
        """

        return self.contexts


# ------------------------------------------------------------------------------
#
class DefaultSession(Session, metaclass=ru.Singleton):

    @rus.takes   ('DefaultSession')
    @rus.returns (rus.nothing)
    def __init__ (self, uid=None):

        # the default session picks up default contexts, from all context
        # adaptors.  To implemented, we have to do some legwork: get the engine,
        # dig through the registered context adaptors, and ask each of them for
        # default contexts.

        super(DefaultSession, self).__init__(default=False, uid=uid)

        _engine = engine.Engine()

        if 'radical.saga.Context' not in _engine._adaptor_registry :
            self._logger.warning ("no context adaptors found")
            return

        for schema   in _engine._adaptor_registry['radical.saga.Context'] :
            for info in _engine._adaptor_registry['radical.saga.Context'][schema] :

                default_ctxs = []

                try :
                    default_ctxs = info['adaptor_instance']._get_default_contexts ()

                except se.SagaException as e :
                    self._logger.debug   ("adaptor %s failed to provide default" \
                                          "contexts: %s" % (info['adaptor_name'], e))
                    continue


                for default_ctx in default_ctxs :

                    try :
                        self.contexts.append (ctx=default_ctx, session=self)
                        self._logger.debug   ("default context [%-20s] : %s" \
                                         %   (info['adaptor_name'], default_ctx))

                    except se.SagaException as e :
                        self._logger.debug   ("skip default context [%-20s] : %s : %s" \
                                         %   (info['adaptor_name'], default_ctx, e))
                        continue


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import radical.saga as rs


class MyContextA (rs.Context) :

    def __init__ (self, ctype) :

        self._apitype  = 'rs.Context'
        super (MyContextA, self).__init__ (ctype)


class MyContextB (rs.Context) :

    def __init__ (self, ctype) :

        self._apitype  = 'rs.Context'
        rs.Context.__init__ (self, ctype)


cs = rs.Context ('ssh') ; print("saga: %s" % cs)
ca = MyContextA ('ssh') ; print("mc a: %s" % ca)
cb = MyContextB ('ssh') ; print("mc b: %s" % cb)





__author__    = "The RADICAL Team"
__copyright__ = "Copyright 2012-2019, The SAGA Project"
__license__   = "MIT"


""" LSF job adaptor implementation
"""

import re
import os
import copy
import time
import threading
import datetime

from urllib.parse import parse_qs

import radical.utils  as ru

from ..    import base
from ..    import cpi as cpi_base
from ..cpi import job as cpi

from ...              import exceptions as rse
from ...              import job        as rsj
from ...utils         import pty_shell  as rsups


SYNC_CALL  = cpi_base.decorators.SYNC_CALL
ASYNC_CALL = cpi_base.decorators.ASYNC_CALL

SYNC_WAIT_UPDATE_INTERVAL = 1  # seconds
MONITOR_UPDATE_INTERVAL   = 3  # seconds

# Intel LSF hosts have SMT default to 4
SMT_DEFAULT = 1
SMT_VALID_VALUES = [1, 2, 4]

# FIXME: will be taken from resource config
RESOURCES = {
    'summit': {'cpn': 42,
               'gpn': 6,
               'valid_alloc_flags': [
                   'gpumps',
                   'gpudefault',
                   'nvme',
                   'spectral',
                   'maximizegpfs'
               ]},
    'lassen': {'cpn': 40,
               'gpn': 4,
               'valid_alloc_flags': [
                   'atsdisable',
                   'autonumaoff',
                   'cpublink',
                   'ipisolate'
               ]}
}


# ------------------------------------------------------------------------------
#
class _job_state_monitor(threading.Thread):
    """
    thread that periodically monitors job states
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, job_service):

        self.logger = job_service._logger
        self.js     = job_service
        self._term  = threading.Event()

        super(_job_state_monitor, self).__init__()
        self.setDaemon(True)


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self.logger.info('stop  thread for %s', self.js.get_url())
        self._term.set()


    # --------------------------------------------------------------------------
    #
    def run(self):

        self.logger.info('start thread for %s', self.js.get_url())
        while not self._term.isSet():
        
            try:
                # do bulk updates here! we don't want to pull information
                # job by job. that would be too inefficient!
                jobs = self.js.jobs
                for job in jobs:

                    # if the job hasn't been started, we can't update its
                    # state. we can tell if a job has been started if it
                    # has a job id
                    job_info = self.js.jobs[job]
                    if not job_info.get('job_id'):
                        continue

                    # we only need to monitor jobs that are not in a
                    # terminal state, so we can skip the ones that are
                    # either done, failed or canceled
                    state = job_info['state']
                    if state in rsj.FINAL:
                        continue

                    # job is not final and we got new info - replace job info
                    new_info = self.js._job_get_info(job)
                    self.js.jobs[job] = new_info

                    # we only care to state updates though when the state
                    # actually changed from last time
                    new_state = new_info['state']
                    if new_state == state:
                        continue

                    # fire job state callback if 'state' has changed
                    self.logger.info("update Job %s (state: %s)"
                                    % (job, new_state))
                    job._api()._attributes_i_set('state', new_state,
                                                 job._api()._UP, True)

                time.sleep(MONITOR_UPDATE_INTERVAL)

            except Exception as e:
                self.logger.exception("job monitoring thread failed")
                break

        self.logger.info('close thread for %s', self.js.get_url())


# ------------------------------------------------------------------------------
#
def _lsf_to_saga_jobstate(lsfjs):
    """
    translates a lsf one-letter state to saga
    """

    if   lsfjs in ['RUN']                     : return rsj.RUNNING
    elif lsfjs in ['WAIT', 'PEND']            : return rsj.PENDING
    elif lsfjs in ['DONE']                    : return rsj.DONE
    elif lsfjs in ['UNKNOWN', 'ZOMBI', 'EXIT']: return rsj.FAILED
    elif lsfjs in ['USUSP', 'SSUSP', 'PSUSP'] : return rsj.SUSPENDED
    else                                      : return rsj.UNKNOWN


# ------------------------------------------------------------------------------
#
def _lsfscript_generator(url, logger, jd, ppn, lsf_version, queue):
    """
    generates an LSF script from a SAGA job description
    """

    lsf_bsubs   = ''
    command     = ''
    env_string  = ''

    if jd.executable: command += "%s " % (jd.executable)
    if jd.arguments : command += ' '.join(jd.arguments)

    bsub_queue = queue or jd.queue
    if bsub_queue          : lsf_bsubs += "#BSUB -q %s \n" % bsub_queue

    if jd.name             : lsf_bsubs += "#BSUB -J %s \n" % jd.name
    if jd.job_contact      : lsf_bsubs += "#BSUB -u %s \n" % jd.job_contact
    if jd.working_directory: lsf_bsubs += "#BSUB -cwd %s \n" \
                                                     %  jd.working_directory
    if jd.wall_time_limit  : lsf_bsubs += "#BSUB -W %s:%s \n" \
                                                     % (int(jd.wall_time_limit / 60),
                                                        int(jd.wall_time_limit % 60))

    # if working directory is set, we want stdout to end up in the
    # working directory as well, unless it contains a specific
    # path name - otherwise we pass `output` as is.
    if jd.output:
        if os.path.isabs(jd.output): path = ''
        elif jd.working_directory  : path = '%s/' % jd.working_directory
        else                       : path = ''
        lsf_bsubs += "#BSUB -o %s%s \n" % (path, jd.output)

    # same holds for stderr
    if jd.error:
        if os.path.isabs(jd.error): path = ''
        elif jd.working_directory : path = '%s/' % jd.working_directory
        else                      : path = ''
        lsf_bsubs += "#BSUB -e %s%s \n" % (path, jd.error)

    if jd.project and ':' in jd.project:
        account, reservation = jd.project.split(':', 1)
        lsf_bsubs += "#BSUB -P %s \n" % account
        lsf_bsubs += "#BSUB -U %s \n" % reservation

    elif jd.project:
        lsf_bsubs += "#BSUB -P %s \n" % jd.project

    # Request enough nodes to cater for the number of gpus and cores requested
    if not jd.total_cpu_count: total_cpu_count = 1
    else                     : total_cpu_count = jd.total_cpu_count

    if not jd.total_gpu_count: total_gpu_count = 1
    else                     : total_gpu_count = jd.total_gpu_count

    hostname = url.host

    if not hostname or 'localhost' in hostname:
        out, _, ret = ru.sh_callout('hostname -f')
        if ret: hostname = os.environ.get('HOSTNAME', '')
        else  : hostname = out.strip()

    if not hostname:
        raise RuntimeError('cannot determine target host f or %s' % url)

    cpn, gpn, smt, valid_alloc_flags = 0, 1, SMT_DEFAULT, []
    for resource_name in RESOURCES:
        if resource_name in hostname:
            smt = jd.system_architecture.get('smt') or smt
            cpn = RESOURCES[resource_name]['cpn'] * smt
            gpn = RESOURCES[resource_name]['gpn']
            valid_alloc_flags = RESOURCES[resource_name]['valid_alloc_flags']
            break

    if not cpn:
        raise ValueError('LSF host (%s) not yet supported' % hostname)

    if smt not in SMT_VALID_VALUES:
        smt = SMT_DEFAULT

    cpu_nodes = int(total_cpu_count / cpn)
    if total_cpu_count > (cpu_nodes * cpn):
        cpu_nodes += 1

    gpu_nodes = int(total_gpu_count / gpn)
    if total_gpu_count > (gpu_nodes * gpn):
        gpu_nodes += 1

    nodes = max(cpu_nodes, gpu_nodes)
    lsf_bsubs += "#BSUB -nnodes %s \n" % str(nodes)

    alloc_flags = []
    for flag in jd.system_architecture.get('options', []):
        if flag.lower() in valid_alloc_flags:
            alloc_flags.append(flag.lower())
    alloc_flags.append('smt%d' % smt)
    lsf_bsubs += "#BSUB -alloc_flags '%s' \n" % ' '.join(alloc_flags)

    env_string += "export RADICAL_SAGA_SMT=%d" % smt
    if jd.environment:
        for k, v in jd.environment.items():
            env_string += " %s=%s" % (k, v)

    # escape double quotes and dollar signs, otherwise 'echo |'
    # further down won't work
    # only escape '$' in args and exe. not in the bsubs
    command   = command.replace('$', '\\$')
    lsfscript = "\n#!/bin/bash \n%s\n%s\n%s" % (lsf_bsubs, env_string, command)
    lsfscript = lsfscript.replace('"', '\\"')

    return lsfscript


# ------------------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0


# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME    = "radical.saga.adaptors.lsfjob"
_ADAPTOR_SCHEMAS = ["lsf", "lsf+ssh", "lsf+gsissh"]


# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes":   [rsj.NAME,
                          rsj.EXECUTABLE,
                          rsj.ARGUMENTS,
                          rsj.ENVIRONMENT,
                          rsj.INPUT,
                          rsj.OUTPUT,
                          rsj.ERROR,
                          rsj.QUEUE,
                          rsj.PROJECT,
                          rsj.WALL_TIME_LIMIT,
                          rsj.WORKING_DIRECTORY,
                          rsj.SPMD_VARIATION,
                          rsj.PROCESSES_PER_HOST,
                          rsj.TOTAL_CPU_COUNT,
                          rsj.TOTAL_GPU_COUNT,
                          rsj.SYSTEM_ARCHITECTURE],
    "job_attributes":    [rsj.EXIT_CODE,
                          rsj.EXECUTION_HOSTS,
                          rsj.CREATED,
                          rsj.STARTED,
                          rsj.FINISHED],
    "metrics":           [rsj.STATE],
    "callbacks":         [rsj.STATE],
    "contexts":          {"ssh"      : "SSH public/private keypair",
                          "x509"     : "GSISSH X509 proxy context",
                          "userpass" : "username/password pair (ssh)"}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name":          _ADAPTOR_NAME,
    "capabilities":  _ADAPTOR_CAPABILITIES,
    "description":  """
The LSF adaptor allows to run and manage jobs on
`LSF <https://en.wikipedia.org/wiki/Platform_LSF>`_ controlled HPC clusters.
""",
    "example": "examples/jobs/lsfjob.py",
    "schemas": {"lsf":        "connect to a local cluster",
                "lsf+ssh":    "connect to a remote cluster via SSH",
                "lsf+gsissh": "connect to a remote cluster via GSISSH"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {"name"        : _ADAPTOR_NAME,
                 "version"     : "v0.2",
                 "schemas"     : _ADAPTOR_SCHEMAS,
                 "capabilities": _ADAPTOR_CAPABILITIES,
                 "cpis": [
                             {
                                 "type": "radical.saga.job.Service",
                                 "class": "LSFJobService"
                             },
                             {
                                 "type": "radical.saga.job.Job",
                                 "class": "LSFJob"
                             }
                         ]
                }


# ------------------------------------------------------------------------------
# The adaptor class
class Adaptor(base.Base):
    '''
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        base.Base.__init__(self, _ADAPTOR_INFO)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.epoch = datetime.datetime(1970,1,1)


    # --------------------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass

    # --------------------------------------------------------------------------
    #
    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise rse.BadParameter("Cannot parse job id '%s'" % id)

        return (match.group(1), match.group(2))


# ------------------------------------------------------------------------------
#
class LSFJobService(cpi.Service):
    """
    implements cpi.job.Service
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        self._mt  = None
        _cpi_base = super(LSFJobService, self)
        _cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor

    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self.close()


    # --------------------------------------------------------------------------
    #
    def close(self):

        if  self.mt:
            self._logger.info("stop   monitoring thread: %s", self.rm)
            self.mt.stop()
            self.mt.join(10)  # don't block forever on join()

        self._logger.info("stopped monitoring thread: %s", self.rm)

        self.finalize(True)


    # --------------------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):

        if  kill_shell:
            if  self.shell:
                self.shell.finalize(True)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """ service instance constructor
        """
        self.rm      = rm_url
        self.session = session
        self.ppn     = 1
        self.queue   = None
        self.shell   = None
        self.jobs    = dict()

        # the monitoring thread - one per service instance
        self.mt = _job_state_monitor(job_service=self)
        self.mt.start()

        rm_scheme = rm_url.scheme
        pty_url   = ru.Url(rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query:
            for key, val in parse_qs(rm_url.query).items():
                if key == 'queue':
                    self.queue = val[0]
                else:
                    raise rse.BadParameter('unsupported url query %s' % key)

        # we need to extrac the scheme for PTYShell. That's basically the
        # job.Serivce Url withou the lsf+ part. We use the PTYShell to execute
        # lsf commands either locally or via gsissh or ssh.
        elems = rm_scheme.split('+')
        if   'gsissh' in elems: pty_url.scheme = "gsissh"
        elif 'ssh'    in elems: pty_url.scheme = "ssh"
        else                  : pty_url.scheme = "fork"

        # these are the commands that we need in order to interact with LSF.
        # the adaptor will try to find them during initialize(self) and bail
        # out in case they are note avaialbe.
        self._commands = {'bqueues':  dict(),
                          'bjobs':    dict(),
                          'bsub':     dict(),
                          'bkill':    dict()}

        self.shell = rsups.PTYShell(pty_url, self.session)

      # self.shell.set_initialize_hook(self.initialize)
      # self.shell.set_finalize_hook(self.finalize)

        self.initialize()
        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # check if all required lsf tools are available
        for cmd in self._commands:

            ret, out, _ = self.shell.run_sync("which %s " % cmd)
            if ret != 0:
                raise rse.NoSuccess("Couldn't find LSF tools: %s" % out)

            else:
                path = out.strip()  # strip removes newline
                ret, out, _ = self.shell.run_sync("%s -V" % cmd)
                if ret != 0:
                    raise rse.NoSuccess("Couldn't find LSF tools: %s" % out)
                else:
                    # version is reported as: "version: x.y.z"
                    version = out.split("\n")[0]

                    # add path and version to the command dictionary
                    self._commands[cmd]["path"]    = path
                    self._commands[cmd]["version"] = version

        self._logger.info("Found LSF tools: %s" % self._commands)

      # # see if we can get some information about the cluster, e.g.,
      # # different queues, number of processes per node, etc.
      # # TODO: this is quite a hack. however, it *seems* to work quite
      # #       well in practice.
      # ret, out, _ = self.shell.run_sync(
      #                   'unset GREP_OPTIONS; %s -a | grep -E "(np|pcpu)"'
      #                   % self._commands['pbsnodes']['path'])
      # if ret != 0:
      #
      #     raise rse.NoSuccess("Error running pbsnodes: %s" % out)
      # else:
      #    # this is black magic. we just assume that the highest occurence
      #    # of a specific np is the number of processors (cores) per compute
      #    # node. this equals max "PPN" for job scripts
      #     ppn_list = dict()
      #     for line in out.split('\n'):
      #         np = line.split(' = ')
      #         if len(np) == 2:
      #             np = np[1].strip()
      #             if np in ppn_list:
      #                 ppn_list[np] += 1
      #             else:
      #                 ppn_list[np] = 1
      #     self.ppn = max(ppn_list, key=ppn_list.get)
      #     self._logger.debug("Found the following 'ppn' configurations: %s. \
      #                         Using %s as default ppn."
      #                       % (ppn_list, self.ppn))


    # --------------------------------------------------------------------------
    #
    def _job_run(self, job_obj):
        """ runs a job via qsub
        """
        # get the job description
        jd = job_obj.jd

        # normalize working directory path
        if  jd.working_directory:
            jd.working_directory = os.path.normpath(jd.working_directory)

        if self.queue and jd.queue:
            self._logger.warning("Job service for queue '%s', \
                                  job goes to queue: '%s'. Using '%s'." %
                                  (self.queue, jd.queue, self.queue))

        try:
            # create an LSF job script from SAGA job description
            script = _lsfscript_generator(url=self.rm, logger=self._logger,
                                jd=jd, ppn=self.ppn,
                                lsf_version=self._commands['bjobs']['version'],
                                queue=self.queue)

            self._logger.info("Generated LSF script: %s" % script)

        except Exception as e:
            raise rse.BadParameter(str(e)) from e

        # try to create the working directory (if defined)
        # WARNING: this assumes a shared filesystem between login node and
        #          compute nodes.
        if jd.working_directory:
            pwd = jd.working_directory
            self._logger.info("Creating working directory %s" % pwd)
            ret, out, _ = self.shell.run_sync("mkdir -p %s" % pwd)

            if ret:
                raise rse.NoSuccess("Couldn't create workdir %s" % out)

        # (1) create a temporary file with 'mktemp' and write the contents of
        #     the generated PBS script into it
        # (2) call 'qsub <tmpfile>' to submit the script to the batch system
        #
        cmdline = \
            "SCRIPTFILE=`mktemp -p $HOME -t SAGA-Python-LSFJobScript.XXXXXX` \
             && echo \"%s\" > $SCRIPTFILE \
             && %s $SCRIPTFILE \
             && rm -f $SCRIPTFILE" % (script, self._commands['bsub']['path'])
        ret, out, _ = self.shell.run_sync(cmdline)

        if ret:
            raise rse.NoSuccess("bsub error: %s [%s]" % (out, cmdline))

        # parse the job id. bsub's output looks like this:
        # Job <901545> is submitted to queue <regular>
        lines = out.split("\n")
        lines = [lines for lines in lines if lines != '']  # remove empty

        self._logger.debug('bsub:\n %s' % '\n'.join(lines))

        lsf_job_id = None
        for line in lines:
            if re.search('Job <.+> is submitted to.+queue', line):
                lsf_job_id = re.findall(r'<(.*?)>', line)[0]
                break

        if not lsf_job_id:
            raise Exception("Failed to detect job id after submission.")

        job_id = "[%s]-[%s]" % (self.rm, lsf_job_id)

        self._logger.info("Submitted LSF job with id: %s" % job_id)

        # update job dictionary
        self.jobs[job_obj]['job_id']    = job_id
        self.jobs[job_obj]['submitted'] = job_id

        # set status to 'pending' and trigger callback
        # a guard is in place to not trigger this twice on state updates
        self.jobs[job_obj]['state'] = rsj.PENDING
        job_obj._api()._attributes_i_set('state',
                      self.jobs[job_obj]['state'], job_obj._api()._UP, True)

        # return the job id
        return job_id


    # --------------------------------------------------------------------------
    #
    def _retrieve_job(self, job_id):
        """ see if we can get some info about a job that we don't
            know anything about
        """
        rm, pid = self._adaptor.parse_id(job_id)

        # bjobs -noheader -o 'stat exec_host exit_code submit_time start_time
        # finish_time command job_name delimiter=","' 344077
        # EXIT,summitdev-login1:summitdev-r0c1n12:summitdev-r0c1n12:
        # summitdev-r0c1n12:summitdev-r0c1n12:summitdev-r0c1n12:
        # summitdev-r0c1n12:summitdev-r0c1n12:summitdev-r0c1n12:
        # summitdev-r0c1n12:summitdev-r0c1n12:summitdev-r0c1n12:
        # summitdev-r0c1n12:summitdev-r0c1n12:summitdev-r0c1n12:
        # summitdev-r0c1n12:summitdev-r0c1n12:summitdev-r0c1n12:
        # summitdev-r0c1n12:summitdev-r0c1n12:summitdev-r0c1n12,
        # 2,Oct 16 11:52,Oct 16 11:52,Oct 16 11:53 L,#!/bin/bash;
        # #BSUB -J saga-test;
        # #BSUB -o examplejob.out;
        # #BSUB -e examplejob.err;
        # #BSUB -W 0:10;
        # #BSUB -q batch;
        # #BSUB -P CSC190SPECFEM;
        # #BSUB -nnodes 1;
        # #BSUB -alloc_flags 'gpumps smt4';
        # export  FILENAME=testfile;/bin/touch \$FILENAME " > $SCRIPTFILE &&
        # /sw/sources/lsf-tools/bin/bsub $SCRIPTFILE && rm -f $SCRIPTFILE,
        # saga-test

        ret, out, _ = self.shell.run_sync("%s -noheader -o 'exit_code stat "
                "exec_host submit_time start_time finish_time job_name command "
                "delimiter=\",\"' %s"
                % (self._commands['bjobs']['path'], pid))

        if ret != 0:
            raise rse.NoSuccess("reconnect error '%s': %s" % (job_id, out))

        # the job seems to exist on the backend. let's gather some data
        job_info = dict()

        results = out.split(',')

        if results[0] != '-': job_info['returncode'] = int(results[0])
        else                : job_info['returncode'] = None

        job_info['state']       = _lsf_to_saga_jobstate(results[1])
        job_info['exec_hosts']  = results[2]
        job_info['create_time'] = results[3]
        job_info['start_time']  = results[4]
        job_info['end_time']    = results[5]
        job_info['name']        = results[6]
        job_info['gone']        = False

        cmd  = results[7]
        exe  = cmd.split()[0]
        args = cmd.split()[1:]

        jd = rsj.Description()
        jd.executable = exe
        jd.arguments  = args
        jd.name       = results[7]

        return [job_info, jd]


    # --------------------------------------------------------------------------
    #
    def _job_get_info(self, job_obj):
        """
        get job attributes via bjob
        """

        # if we don't have the job in our dictionary, we don't want it
        if job_obj not in self.jobs:
            raise rse.NoSuccess("Unknown job %s" % job_obj._id)

        # prev. info contains the info collect when _job_get_info
        # was called the last time
        prev_info = self.jobs[job_obj]

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever
        if prev_info['gone'] is True:
            return prev_info

        # curr. info will contain the new job info collect. it starts off
        # as a copy of prev_info (don't use deepcopy because there is an API
        # object in the dict -> recursion)

        curr_info = copy.deepcopy(prev_info)

        rm, pid = self._adaptor.parse_id(job_obj._id)

        # run the LSF 'bjobs' command to get some infos about our job
        # the result of bjobs <id> looks like this:
        #
        # JOBID USER  STAT QUEUE   FROM_HOST   EXEC_HOST  JOB_NAME  SUBMIT_TIME
        # 90154 oweid DONE regular yslogin5-ib ys3833-ib  *FILENAME Nov 11 12:06
        #
        # If we add the -nodeader flag, the first row is ommited

        ret, out, _ = self.shell.run_sync("%s -noheader %s"
                                       % (self._commands['bjobs']['path'], pid))

        if not ret:
            # parse the result
            results = out.split()
            curr_info['state']      = _lsf_to_saga_jobstate(results[2])
            curr_info['exec_hosts'] = results[5]

        else:

            if "Illegal job ID" not in out:
                raise rse.NoSuccess("bjobs error: %s" % out)

            # Let's see if the previous job state was running or pending. in
            # that case, the job is gone now, which can either mean DONE,
            # or FAILED. the only thing we can do is set it to 'DONE'
            curr_info['gone'] = True
            self._logger.warning("job disappeared - set to DONE")

            if prev_info['state'] in [rsj.RUNNING, rsj.PENDING]:
                curr_info['state'] = rsj.DONE
            else:
                curr_info['state'] = rsj.FAILED

        # return the new job info dict
        return curr_info


    # --------------------------------------------------------------------------
    #
    def _job_get_state(self, job_obj):

        return self.jobs[job_obj]['state']


    # --------------------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_obj):

        ret = self.jobs[job_obj]['returncode']

        # FIXME: 'None' should cause an exception
        if ret is None: return None
        else          : return int(ret)


    # --------------------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_obj):

        return self.jobs[job_obj]['exec_hosts']


    # --------------------------------------------------------------------------
    #
    def _job_get_create_time(self, job_obj):

        return self.jobs[job_obj]['create_time']


    # --------------------------------------------------------------------------
    #
    def _job_get_start_time(self, job_obj):

        return self.jobs[job_obj]['start_time']


    # --------------------------------------------------------------------------
    #
    def _job_get_end_time(self, job_obj):
        # FIXME: convert to EPOCH
        return self.jobs[job_obj]['end_time']


    # --------------------------------------------------------------------------
    #
    def _job_cancel(self, job_obj):
        """
        cancel the job via 'qdel'
        """

        rm, pid     = self._adaptor.parse_id(job_obj._id)
        ret, out, _ = self.shell.run_sync("%s %s\n"
                                       % (self._commands['bkill']['path'], pid))

        if ret:
            raise rse.NoSuccess("qdel error: %s" % out)

        # assume the job was succesfully canceled
        self.jobs[job_obj]['state'] = rsj.CANCELED


    # --------------------------------------------------------------------------
    #
    def _job_wait(self, job_obj, timeout):
        """
        wait for the job to finish or fail
        """

        time_start = time.time()
        rm, pid    = self._adaptor.parse_id(job_obj._id)

        while True:

            state = self.jobs[job_obj]['state']  # this gets updated in the bg.

            if state in [rsj.DONE, rsj.FAILED, rsj.CANCELED]:
                return True

            # avoid busy poll
            time.sleep(SYNC_WAIT_UPDATE_INTERVAL)

            # check if we hit timeout
            if timeout >= 0:
                if time.time() - time_start > timeout:
                    return False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """
        implements cpi.job.Service.create_job()
        """

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       False
                         }

        # create a new job object
        job_obj = rsj.Job(_adaptor=self._adaptor,
                         _adaptor_state=adaptor_state)

        # add job to internal list of known jobs.
        self.jobs[job_obj._adaptor] = {
            'state':        rsj.NEW,
            'job_id':       None,
            'exec_hosts':   None,
            'returncode':   None,
            'create_time':  None,
            'start_time':   None,
            'end_time':     None,
            'gone':         False,
            'submitted':    False
        }

        return job_obj


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, jobid):
        """ Implements cpi.job.Service.get_job()
        """

        # try to get some information about this job
        job_info, job_desc = self._retrieve_job(jobid)


        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         # TODO: fill job description
                         "job_description": job_desc,
                         "job_schema":      self.rm.schema,
                         "reconnect":       True,
                         "reconnect_jobid": jobid
                         }

        job = rsj.Job(_adaptor=self._adaptor,
                     _adaptor_state=adaptor_state)

        # throw it into our job dictionary.
        self.jobs[job._adaptor] = job_info
        return job


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """
        implements cpi.job.Service.get_url()
        """

        return self.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """
        implements cpi.job.Service.list()
        """

        ids = list()
        ret, out, _ = self.shell.run_sync("%s -a"
                                         % self._commands['bjobs']['path'])

        if ret != 0 and len(out) > 0:
            # ignore errors with no output (no job found)
            # FIXME: check stderr
            raise rse.NoSuccess("bjobs error: %s" % out)

        if not ret:
            # call succeeded, output looks like this:
            # 112059.svc.uc.futuregrid testjob oweidner 0 Q batch
            # 112061.svc.uc.futuregrid testjob oweidner 0 Q batch
            for line in out.split("\n"):
                if len(line.split()) > 1:
                    elems = line.split()[0].split('.')
                    jobid = "[%s]-[%s]" % (self.rm, elems[0])
                    ids.append(str(jobid))

        return ids


    # --------------------------------------------------------------------------
    #
    def container_run(self, jobs):

        # TODO: this is not optimized yet
        for job in jobs:
            job.run()


    # --------------------------------------------------------------------------
    #
    def container_wait(self, jobs, mode, timeout):

        if timeout:
            raise rse.NoSuccess("bulk wait timeout is not implemented")

        # TODO: this is not optimized yet
        for job in jobs:
            job.wait()


    # --------------------------------------------------------------------------
    #
    def container_cancel(self, jobs, timeout):

        if timeout:
            raise rse.NoSuccess("bulk cancel timeout is not implemented")

        # TODO: this is not optimized yet
        for job in jobs:
            job.cancel()


# ------------------------------------------------------------------------------
#
class LSFJob(cpi.Job):
    """
    implements cpi.job.Job
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        # initialize parent class
        _cpi_base = super(LSFJob, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    def _get_impl(self):
        return self


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, job_info):
        """ implements cpi.job.Job.init_instance()
        """
        # init_instance is called for every new Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        if job_info['reconnect'] is True:
            self._id      = job_info['reconnect_jobid']
            self._name    = self.jd.get(rsj.NAME)
            self._started = True
        else:
            self._id      = None
            self._name    = self.jd.get(rsj.NAME)
            self._started = False

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements cpi.job.Job.get_state()
        """
        return self.js._job_get_state(job_obj=self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """ implements cpi.job.Job.wait()
        """
        if not self._started:
            raise rse.IncorrectState("job has not been started")

        self.js._job_wait(job_obj=self, timeout=timeout)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """
        implements cpi.job.Job.cancel()
        """

        if not self._started:
            raise rse.IncorrectState("job has not been started")

        self.js._job_cancel(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """
        implements cpi.job.Job.run()
        """

        self._id = self.js._job_run(self)
        self._started = True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """
        implements cpi.job.Job.get_service_url()
        """

        return self.js.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """
        implements cpi.job.Job.get_id()
        """

        return self._id


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name(self):
        """
        Implements cpi.job.Job.get_name()
        """

        return self._name


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """
        implements cpi.job.Job.get_exit_code()
        """

        if not self._started:
            return None

        return self.js._job_get_exit_code(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """
        implements cpi.job.Job.get_created()
        """

        if not self._started:
            return None

        return self.js._job_get_create_time(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """
        implements cpi.job.Job.get_started()
        """

        if not self._started:
            return None

        return self.js._job_get_start_time(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """
        implements cpi.job.Job.get_finished()
        """

        if not self._started:
            return None

        # FIXME: convert to EPOCH
        return self.js._job_get_end_time(self)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """
        implements cpi.job.Job.get_execution_hosts()
        """

        if not self._started:
            return None

        return self.js._job_get_execution_hosts(self)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" This examples shows how to run groups of jobs using the
    'local' file adaptor. This example uses job containers for
    simplified and optimized bulk job handling.

    Job container can be used to easily model dependencies between
    groups of different jobs, e.g., in workflow scenarios. In this example,
    we execute 'num_job_groups' containers of jobs_per_group' number of
    parallel jobs sequentially::

      C1[j1,j2,j3,j4,...] -> C2[j1,j2,j3,j4,...] -> C3[j1,j2,j3,j4,...] -> ...

    Depending on the adaptor implementation, using job containers can be
    quite advantageous in terms of call latency. Some adaptors implement
    special bulk operations for container management, which makes them
    generally much faster than iterating over and operating on individual jobs.
"""

import sys
import random
import time

import radical.saga as rs


URL = "condor+gsissh://xd-login.opensciencegrid.org"
URL = "fork://locahost/"

def main():

    # number of job 'groups' / containers
    num_job_groups = 2
    # number of jobs per container
    jobs_per_group = 10

    try:
        # all jobs in this example are running on the same job service
        # this is not a requirement though. s
        service = rs.job.Service(URL)
        print(service.url)

        t1 = time.time()
        # create and populate our containers
        containers = list()
        for c in range(0, num_job_groups):
            # create containers
            containers.append(rs.job.Container())
            for j in range(0, jobs_per_group):
                # add jobs to container. to make things a bit more
                # interesting, we give each job a random runtime (1-60s)
                jd = rs.job.Description()
                jd.environment = {'RUNTIME': random.randrange(10, 60)}
                jd.executable  = '/bin/sleep'
                jd.arguments   = ['$RUNTIME']
                jd.name        = ['job.%02d.%03d' % (c, j)]
                j = service.create_job(jd)
                containers[c].add(j)

        # execute the containers sequentially
        for c in range(0, num_job_groups):
            print('Starting container %s ... ' % c)
            containers[c].run()

            for j in containers[c].get_tasks():
                print('%s: %s: %s' % (j.name, j.id, j.state))

            print(containers[c].get_states ())
            containers[c].cancel()
            containers[c].wait()
            print(containers[c].get_states ())

            # # at this point, all jobs in the container
            # # have finished running. we can now print some statistics
            # for job in containers[c].jobs:
            #     print "  * Job id=%s state=%s rc=%s exec_host=%s start_time=%s end_time=%s" \
            #       % (job.id, job.state, job.exit_code, job.execution_hosts,
            #          job.started, job.finished)
        t2 = time.time()
        print(t2-t1)

        service.close()
        return 0

    except rs.SagaException as ex:
        print("An exception occured: %s " % ((str(ex))))
        # get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" *** %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"


'''This examples shows how to use the rs.Filesystem API
   with the SFTP file adaptor.

   If something doesn't work as expected, try to set
   SAGA_VERBOSE=3 in your environment before you run the
   script in order to get some debug output.

   If you think you have encountered a defect, please 
   report it at: https://github.com/radical-cybertools/radical.saga/issues
'''

import sys

import radical.saga as rs


def main():

    try:
        # Your ssh identity on the remote machine.
        ctx = rs.Context("ssh")

        # Change e.g., if you have a differnent username on the remote machine
        #ctx.user_id = "your_ssh_username"

        session = rs.Session()
        session.add_context(ctx)

        # open home directory on a remote machine
        remote_dir = rs.filesystem.Directory('sftp://localhost/',
                                               session=session)

        # copy .bash_history to /tmp/ on the local machine
        remote_dir.copy('/etc/hosts', 'file://localhost/tmp/')

        # list 'h*' in local /tmp/ directory
        local_dir = rs.filesystem.Directory('file://localhost/tmp/')
        for entry in local_dir.list(pattern='h*'):
            print(entry)
        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import decorators as cpi_dec
from .. import namespace  as cpi_ns

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# keep order of inheritance!  super() below uses MRO
class File (cpi_ns.entry.Entry) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsentry = super  (File, self)
        self._cpi_nsentry.__init__ (api, adaptor)


    @SYNC
    def init_instance        (self, url, flags, session)      : pass
    @ASYNC
    def init_instance_async  (self, url, flags, session)      : pass


    # ----------------------------------------------------------------
    #
    # filesystem methods
    #
    @SYNC
    def is_file_self         (self, ttype)                    : pass
    @ASYNC
    def is_file_self_async   (self, ttype)                    : pass

    @SYNC
    def get_size_self        (self, ttype)                    : pass
    @ASYNC
    def get_size_self_async  (self, ttype)                    : pass

    @SYNC
    def read                 (self, size, ttype)              : pass
    @ASYNC
    def read_async           (self, size, ttype)              : pass

    @SYNC
    def write                (self, data, ttype)              : pass
    @ASYNC
    def write_async          (self, data, ttype)              : pass

    @SYNC
    def seek                 (self, off, whence, ttype)       : pass
    @ASYNC
    def seek_async           (self, off, whence, ttype)       : pass

    @SYNC
    def read_v               (self, iovecs, ttype)            : pass
    @ASYNC
    def read_v_async         (self, iovecs, ttype)            : pass

    @SYNC
    def write_v              (self, data, ttype)              : pass
    @ASYNC
    def write_v_async        (self, data, ttype)              : pass

    @SYNC
    def size_p               (self, pattern, ttype)           : pass
    @ASYNC
    def size_p_async         (self, pattern, ttype)           : pass

    @SYNC
    def read_p               (self, pattern, ttype)           : pass
    @ASYNC
    def read_p_async         (self, pattern, ttype)           : pass

    @SYNC
    def write_p              (self, pattern, data, ttype)     : pass
    @ASYNC
    def write_p_async        (self, pattern, data, ttype)     : pass

    @SYNC
    def modes_e              (self, ttype)                    : pass
    @ASYNC
    def modes_e_async        (self, ttype)                    : pass

    @SYNC
    def size_e               (self, emode, spec, ttype)       : pass
    @ASYNC
    def size_e_async         (self, emode, spec, ttype)       : pass

    @SYNC
    def read_e               (self, emode, spec, ttype)       : pass
    @ASYNC
    def read_e_async         (self, emode, spec, ttype)       : pass

    @SYNC
    def write_e              (self, emode, spec, data, ttype) : pass
    @ASYNC
    def write_e_async        (self, emode, spec, data, ttype) : pass








__author__    = "Andre Merzky, Ole Weidner, Alexander Grill"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils               as ru
import radical.utils.signatures    as rus

from .constants  import *
from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base       as sab
from ..namespace import directory  as nsdir

from .. import session             as ss
from .. import task                as st


# ------------------------------------------------------------------------------
#
class Directory (nsdir.Directory) :
    """
    Represents a (remote) directory.

    The saga.filesystem.Directory class represents, as the name indicates,
    a directory on some (local or remote) filesystem.  That class offers
    a number of operations on that directory, such as listing its contents,
    copying files, or creating subdirectories::

        # get a directory handle
        dir = saga.filesystem.Directory("sftp://localhost/tmp/")

        # create a subdir
        dir.make_dir ("data/")

        # list contents of the directory
        files = dir.list ()

        # copy *.dat files into the subdir
        for f in files :
            if f ^ '^.*\.dat$' :
                dir.copy (f, "sftp://localhost/tmp/data/")
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=READ, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        """
        __init__(url, flags=READ, session)

        Construct a new directory object

        :param url:     Url of the (remote) directory
        :type  url:     :class:`saga.Url`

        :param flags:   :ref:`filesystemflags`
        :param session: :class:`saga.Session`

        The specified directory is expected to exist -- otherwise
        a DoesNotExist exception is raised.  Also, the URL must point to
        a directory (not to a file), otherwise a BadParameter exception is
        raised.

        Example::

            # open some directory
            dir = saga.filesystem.Directory("sftp://localhost/tmp/")

            # and list its contents
            files = dir.list ()

        """

        # param checks
        if  not flags : flags = 0
        url = ru.Url (url)

        if  not url.schema :
            url.schema = 'file'

        if  not url.host :
            url.host = 'localhost'

        self._nsdirec = super  (Directory, self)
        self._nsdirec.__init__ (url, flags, session,
                                _adaptor, _adaptor_state, _ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create (cls, url=None, flags=READ, session=None, ttype=None) :
        """
        url:       saga.Url
        flags:     saga.replica.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        """

        if  not flags : flags = 0
        _nsdir = super (Directory, cls)
        return _nsdir.create (url, flags, session, ttype=ttype)

    # ----------------------------------------------------------------
    #
    # filesystem directory methods
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (('File', st.Task))
    def open (self, path, flags=READ, ttype=None) :
        """
        open(path, flags=READ)

        Open a file in the directory instance namespace. Returns
        a new file object.

        :param path:     The name/path of the file to open
        :type path:      str()
        :param flags:    :ref:`filesystemflags`
        """
        if  not flags : flags = 0

        return self._adaptor.open (path, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (('Directory', st.Task))
    def open_dir (self, path, flags=READ, ttype=None) :
        """
        open_dir(path, flags=READ)

        Open a directory in the directory instance namespace. Returns
        a new directory object.

        :param path:     The name/path of the directory to open
        :type path:      str()
        :param flags:    :ref:`filesystemflags`

        Example::

            # create a subdir 'data' in /tmp
            dir = saga.namespace.Directory("sftp://localhost/tmp/")
            data = dir.open_dir ('data/', saga.namespace.Create)
        """
        if  not flags : flags = 0

        return self._adaptor.open_dir (path, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def get_size (self, path=None, ttype=None) :
        """
        get_size(path=None)

        Return the size of the directory itself or the entry pointed to by `path`.

        :param path:     (Optional) name/path of an entry
        :type path:      str()

        Returns the size of a file or directory (in bytes)

        Example::

            # inspect a file for its size
            dir  = saga.filesystem.Directory("sftp://localhost/tmp/")
            size = dir.get_size ('data/data.bin')
            print(size)
        """
        if path   :  return self._adaptor.get_size      (path, ttype=ttype)
        else      :  return self._adaptor.get_size_self (      ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional (bool))
    @rus.returns (st.Task)
    def close     (self, kill=True, ttype=None) :
        '''
        kill :    bool
        ttype:    saga.task.type enum
        ret:      string / bytearray / saga.Task
        '''
        return self._adaptor.close ()

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_file (self, path=None, ttype=None) :
        """
        is_file(path=None)

        Returns `True` if entry points to a file, `False` otherwise. If `path`
        is not none, the entry pointed to by `path` is inspected instead of the
        directory object itself.

        :param path:     (Optional) name/path of an entry
        :type path:      str()
        """
        if path   :  return self._adaptor.is_file      (path, ttype=ttype)
        else      :  return self._adaptor.is_file_self (      ttype=ttype)


    size  = property (get_size)  # int


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Andre Merzky, Mark Santcroos"
__copyright__ = "Copyright 2015, The SAGA Project"
__license__   = "MIT"

'''This examples shows how to use the rs.Filesystem API
   with the Globus Online file adaptor.

   If something doesn't work as expected, try to set
   SAGA_VERBOSE=3 in your environment before you run the
   script in order to get some debug output.

   If you think you have encountered a defect, please 
   report it at: https://github.com/radical-cybertools/radical.saga/issues
'''

import sys
import os

import radical.saga as rs


def main():

    try:
        ctx = rs.Context("x509")
        ctx.user_proxy = '/Users/mark/proj/myproxy/xsede.x509'

        session = rs.Session()
        session.add_context(ctx)

        source = "go://marksant#netbook/Users/mark/tmp/go/"
        #destination = "go://xsede#stampede/~/tmp/"
        #destination = "go://gridftp.stampede.tacc.xsede.org/~/tmp/"
        destination = "go://oasis-dm.sdsc.xsede.org/~/tmp/"
        #destination = "go://ncsa#BlueWaters/~/tmp/"
        filename = "my_file"

        # open home directory on a remote machine
        source_dir = rs.filesystem.Directory(source)

        # copy .bash_history to /tmp/ on the local machine
        source_dir.copy(filename, destination)

        # list 'm*' in local /tmp/ directory
        dest_dir = rs.filesystem.Directory(destination)
        for entry in dest_dir.list(pattern='%s*' % filename[0]):
            print(entry)

        dest_file = rs.filesystem.File(os.path.join(destination, filename))
        assert dest_file.is_file() == True
        assert dest_file.is_link() == False
        assert dest_file.is_dir() == False
        print('Size: %d' % dest_file.get_size())

        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


'''
A collection of utilities which maps a namespace structure to redis nosql
keys.  It mirrors the Python ``os`` API, to some extent.

The redis key layout is like the following::

    # namespace tree
    /node:/                 : { type: DIR,   ... }
    /node:/etc/             : { type: DIR,   ... }
    /node:/etc/passwd       : { type: ENTRY, ... }
    /node:/etc/pass         : { type: LINK,  tgt : /etc/passwd, ... }
    ...

    # attribute storage for ns entries, aka adverts (incl. object storage)
    /data:/etc/passwd       : { _obj: <serial>,
                               key_1: val_1, key_2 : val_2, ... }
    ...

    # index for ls type ops
    /kids:/                 : [etc, usr, lib, ...]
    /kids:/etc/             : [init.d, rc.0, rc.1, ...
    ...

    # the two below are only enabled on 'attribute_indexing = True
    # index for attribute key lookups
    /keys:/etc/passwd:key_1 : [/vals/etc/passwd, ...]
    /keys:/etc/passwd:key_2 : [/vals/etc/passwd, ...]

    # index for attribute value lookups
    /vals:/etc/passwd:val_1 : [/keys/etc/passwd, ...]
    /vals:/etc/passwd:val_2 : [/keys/etc/passwd, ...]

all wildcard lookup versions will be slow -- only solution (iiuc) would be to
blow the indexes to cover for wildcard expansion - which is always incomplete
anyway...

All parts of the above structure are set/get in a single pipelined transactional
'MULTI' block.  One could consider to move the ops into a lua script, if
performance is insufficient.

TODO:
    - use locks to make thread safe(r)

'''

import redis

import re
import os
import time
import threading as mt

import radical.utils as ru

from ...       import exceptions as rse
from ...advert import constants  as rsac

from . import redis_cache


TYPE   = 'type'
OBJECT = 'object'

DIR    = 'dir'
ENTRY  = 'entry'
LINK   = 'link'

NODE   = 'node'
DATA   = 'data'
KIDS   = 'kids'
KEYS   = 'keys'
VALS   = 'vals'

MON    = 'saga-advert-events'


# ------------------------------------------------------------------------------
#
# for some reason, POSIX allows two leading slashes, but we need path names to
# be unique (they are used as keys...)
def redis_ns_parent(path):

    if path in ['/', '//']:
        return '/'

    return os.path.split(path)[0]


def redis_ns_name(path):

    if path in ['/', '//']:
        return '/'

    return os.path.split(path)[1]


# ------------------------------------------------------------------------------
#
class redis_ns_monitor(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, r, pub):

        self.r      = r
        self.pub    = pub
        self.logger = r.logger

        self.pat = {
                'ATTRIBUTE': re.compile('\s*\[(?P<key>[^=]+)=(?P<val>.+)]\s*')
        }

        self._term   = mt.Event()
        self._worker = mt.Thread(target=self.work)
        self._worker.daemon = True

    # --------------------------------------------------------------------------
    #
    def start(self):

        self._worker.start()

    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._term.set()

    # --------------------------------------------------------------------------
    #
    def work(self):

        while not self._term.is_set():
            self._work()

    # --------------------------------------------------------------------------
    #
    def _work(self):

        try:
            callbacks = self.r.callbacks
            sub       = self.pub.listen()

            while sub:

                info = next(sub)
                data = info['data']

                if not isinstance(data, str):
                    self.logger.warn("ignoring event: %s"  %  data)
                    continue

                # FIXME: need proper regex parsing
                elems = data.split()
                if not len(elems) == 3:
                    self.logger.warn("ignoring event args: %s"  %  data)
                    continue

                self.logger.debug("sub %s"  %  data)
                event = elems[0]
                path  = elems[1]
                args  = elems[2:]

                if path in callbacks:

                    if event == 'ATTRIBUTE':

                        # args are formatted like '[key=val]'
                        match = self.pat[event].match(' '.join(args))

                        if not match:
                            self.logger.warn("parse error for %s" % args)
                            continue

                        # FIXME: error check
                        key     = match.group('key')
                        val     = match.group('val')

                        if key in callbacks[path]:
                            for idx in callbacks[path][key]:

                              # cb  = callbacks[path][key][idx][0]
                                obj = callbacks[path][key][idx][1]

                                obj.set_attribute(key, val, obj._UP)

                    if event == 'ATTRIBUTES':
                        self.logger.warn("unknown event type %s" % event)
                        pass

        except Exception:
            self.logger.exception("monitoring thread died, callback disabled")
            return


# ------------------------------------------------------------------------------
#
class redis_ns_server(redis.Redis):

    def __init__(self, url):

        if url.scheme != 'redis':
            raise rse.BadParameter("unsupported url scheme (%s)" %  url)

        self.url       = url
        self.host      = 'localhost'
        self.port      = 6379
        self.db        = 0
        self.password  = None
        self.errors    = 'strict'

        if url.host    : self.host     = url.host
        if url.port    : self.port     = url.port
        if url.username: self.username = url.username
        if url.password: self.password = url.password

        # create redis client
        t1 = time.time()
        redis.Redis.__init__(self,
                             host      = self.host,
                             port      = self.port,
                             db        = self.db,
                             password  = self.password,
                             errors    = self.errors)
        t2 = time.time()

        # add a logger
        self.logger = ru.Logger('radical.saga')

        # create a cache dict and attach to redis client instance.  Cache
        # lifetime is set to 10 times the redis-connect latency.
        self.cache = redis_cache.Cache(logger=self.logger, ttl=((t2-t1)*10))

        # create a second client to manage the (blocking)
        # pubsub communication for event notifications
        self.r2 = redis.Redis(host      = self.host,
                              port      = self.port,
                              db        = self.db,
                              password  = self.password,
                              errors    = self.errors)

        # set up pubsub endpoint, and start a thread to monitor channels
        self.callbacks = dict()
        self.pub = self.r2.pubsub()
        self.pub.subscribe(MON)
        # FIXME: create one pubsub channel per path (for paths which have
        #        callbacks registered)

        self.monitor = redis_ns_monitor(self, self.pub)
        self.monitor.start()

    def __del__(self):

        if self.pub:
            self.pub.unsubscribe(MON)


# ------------------------------------------------------------------------------
#
class redis_ns_entry:

    # --------------------------------------------------------------------------
    #
    def __init__(self, r, path):

        self.r         = r
        self.path      = path
        self.node      = {TYPE: None}
        self.data      = {}
        self.kids      = []
        self.valid     = False  # not initialized
        self.logger    = r.logger
        self.cache     = r.cache
        self.callbacks = r.callbacks

    # --------------------------------------------------------------------------
    #
    def _dump(self):

        print("self.r        : %s" % str(self.r        ))
        print("self.path     : %s" % str(self.path     ))
        print("self.node     : %s" % str(self.node     ))
        print("self.data     : %s" % str(self.data     ))
        print("self.kids     : %s" % str(self.kids     ))
        print("self.valid    : %s" % str(self.valid    ))
        print("self.logger   : %s" % str(self.logger   ))
        print("self.cache    : %s" % str(self.cache    ))
        print("self.callbacks: %s" % str(self.callbacks))

    # --------------------------------------------------------------------------
    #
    @classmethod
    def opendir(self, r, path, flags):

        r.logger.debug("redis_ns_entry.opendir %s" % path)

        ns_en = redis_ns_entry(r, path)

        try:
            ns_en.fetch()
        except Exception as e:

            if rsac.CREATE         & flags or \
               rsac.CREATE_PARENTS & flags    :
                ns_en = redis_ns_entry(r, path)
                ns_en.mkdir(flags)
            else:
                raise rse.BadParameter(
                    "Cannot open %s (does not exist)" % path) from e

        if not ns_en.is_dir():
            raise rse.BadParameter("Cannot open %s (not a directory)" % path)

        return ns_en

    # --------------------------------------------------------------------------
    #
    @classmethod
    def open(self, r, path, flags):
        # FIXME: the checks below make open quite slow, as we travel down the
        #        path components.  This should be done in a single pipeline.

        r.logger.debug("redis_ns_entry.open %s" % path)

        # make sure parent dir exists
        try:
            parent_path = redis_ns_parent(path)
            # go down the rabbit hole
            parent = self.opendir(r, parent_path, flags)
        except Exception as e:
            raise rse.BadParameter(
                "Cannot open parent %s (%s)" % (path, e)) from e

        # try to open entry itself
        ns_en = redis_ns_entry(r, path)
        try:
            ns_en.fetch()
        except Exception as e:

            if rsac.CREATE & flags:
                ns_en.node[TYPE] = ENTRY
                ns_en.create(flags)
            else:
                raise rse.BadParameter(
                    "Cannot open %s (no such entry)" % path) from e

        return ns_en

    # --------------------------------------------------------------------------
    #
    def mkdir(self, flags):
        '''
        Don't call this  directly -- to create a dir, call opendir with
        'create'/'create_parents' and 'exclusive'.  If called, assumes that
        entry is invalid (and thus does not yet exist).
        '''

        path = self.path

        if self.valid:
            raise rse.IncorrectState("mkdir %s failed, entry exists" % path)

        self.logger.debug("redis_ns_entry.mkdir %s" % path)

        # if / does not exist, we always create it - no need for checks
        if path != '/':

            # if CREATE_PARENTS is set, we need to check all parents, and need
            # to create them as needed.  We go top down, and terminate once
            # a parent is found
            if rsac.CREATE_PARENTS & flags:

                # this will recursively travel down the chimney hole, and stop
                # whenever it finds an existing  directory
                parent = redis_ns_parent(path)
                pe = redis_ns_entry.opendir(self.r, parent, rsac.CREATE_PARENTS)

            else:
                # if 'CREATE_PARENTS is not set, parent must exist.
                parent = redis_ns_parent(path)
                pe     = None

                try:
                    pe = redis_ns_entry(self.r, parent)
                    pe.fetch()

                except Exception as e:
                    raise rse.BadParameter("parent does not exist on %s: %s" %
                                           (path, e)) from e

                if not pe.is_dir():
                    raise rse.BadParameter("parent is no directory for %s: %s" %
                                           (path, parent))

        self.node[TYPE] = DIR
        self.create(flags)

    # --------------------------------------------------------------------------
    #
    def create(self, flags=0):
        '''
        This assumes that the target entry does not exist.  If flags contains
        CREATE though, we should not raise an exception if it in fact does not.
        '''

        path = self.path

        if self.valid:
            raise rse.IncorrectState("mkdir on %s fails: already exists" % path)

        # FIXME: need to ensure this via a WATCH call.

        self.logger.debug("redis_ns_entry.create %s" % path)

        name   = redis_ns_name(path)
        parent = redis_ns_parent(path)

        now    = time.time()

        self.node['mtime'] = now
        self.node['ctime'] = now

        p = self.r.pipeline()
        # FIXME: add guard
        if len(self.node): p.hset(NODE + ':' + path, None, None, self.node)
        if len(self.data): p.hset(DATA + ':' + path, None, None, self.data)
        if len(self.kids): p.hset(KIDS + ':' + path, None, None, self.kids)

        # add entry as  kid to parent
        # FIXME: avoid duplicated entries!
        if path != '/':
            p.sadd(KIDS + ':' + parent, path)

        # add new index entries
        for key in self.data:
            val = self.data[key]
            p.sadd(KEYS + ':' + str(key), path)
            p.sadd(VALS + ':' + str(val), path)

        # FIXME: eval vals
        p.execute()

        # issue notification about entry creation to parent dir
        self.logger.debug("pub CREATE %s [%s]" % (parent, name))
        self.r.publish(MON,   "CREATE %s [%s]" % (parent, name))

        # refresh cache state
        self.cache.set(NODE + ':' + path, self.node)
        self.cache.set(DATA + ':' + path, self.data)
        self.cache.set(KIDS + ':' + path, self.kids)

        self.valid = True

    # --------------------------------------------------------------------------
    #
    def __str__(self):

        return "[%-4s] %-25s %s %s : %s" \
             % (self.node[TYPE], self.path, self.node, self.kids, self.data)

    # --------------------------------------------------------------------------
    #
    def is_dir(self):

        if self.node[TYPE] == DIR:
            return True
        else:
            return False

    # --------------------------------------------------------------------------
    #
    def list(self):

        if not self.node[TYPE] == DIR:
            raise rse.IncorrectState("list() only supported on directories")

        self.fetch()

        return self.kids

    # --------------------------------------------------------------------------
    #
    def fetch(self):

        self.logger.debug("redis_ns_entry.fetch %s" % self.path)

        path = self.path

        try:
            self.node = self.cache.get(NODE + ':' + path)
            self.data = self.cache.get(DATA + ':' + path)

            if self.node[TYPE] == DIR:
                self.kids = self.cache.get(KIDS + ':' + path)

            self.valid = True
            return

        except Exception:
            # some cache ops failed, so we need to properly fetch data.  We
            # simply fetch all of it
            pass

        try:
            p = self.r.pipeline()
            p.hgetall (NODE + ':' + path)
            p.hgetall (DATA + ':' + path)
            p.smembers(KIDS + ':' + path)
            values = p.execute()

            if len(values) != 3:
                self.valid = False
                return

            self.valid = True

            # FIXME: check val types
            self.node = values[0]
            self.data = values[1]
            self.kids = values[2]  # will be 'None' for non-DIR entries

            if len(self.node) == 0:
                self.valid = False
                raise rse.IncorrectState("backend entry gone or corrupted")

            # cache our newly found entries
            self.cache.set(NODE + ':' + path, self.node)
            self.cache.set(DATA + ':' + path, self.data)
            self.cache.set(KIDS + ':' + path, self.kids)

            # fetched from redis ok
            self.valid = True

        except Exception as e:
            self.valid = False
            raise rse.IncorrectState(
                "backend entry gone or corrupted: %s" % e) from e

    # --------------------------------------------------------------------------
    #
    def get_data(self):

        self.logger.debug("redis_ns_entry.get_data %s" % self.path)

        self.fetch()  # refresh cache/state as needed

        return self.data

    # --------------------------------------------------------------------------
    #
    def get_object(self):

        self.logger.debug("redis_ns_entry.get_object %s" % self.path)

        self.fetch()  # refresh cache/state as needed

        # FIXME: dig '_object' out of the data hash, and de-serialize it to the
        # respective SAGA API object (if possible)

        return None

    # --------------------------------------------------------------------------
    #
    def get_key(self, key):

        self.logger.debug("redis_ns_entry.get_key %s" % (key))

        self.fetch()  # refresh cache/state as needed

        if key not in self.data:
            raise rse.BadParameter("no such attribute (%s)" %  key)

        val = self.data[key]

        if key == 'foo' and val == 'stop':
            self.cache._dump()

        return val

    # --------------------------------------------------------------------------
    #
    def set_key(self, key, val):

        path = self.path
        self.logger.debug("set_key %s: %s" % (path, key))

        self.fetch()  # refresh cache/state as needed

        # FIXME: we only fetch() for the indexes - we should optimize that again
        # by moving index consolidation into a separate thread (p.srem below)

        if key in self.data and self.data[key] == val:

            # nothing changed - so just trigger the set event
            self.logger.debug("Pub ATTRIBUTE %s [%s=%s]" % (path, key, val))
            self.r.publish(MON,   "ATTRIBUTE %s [%s=%s]" % (path, key, val))

            # nothing else to do
            return

        # need to set the key, and update the key/val indexes
        p = self.r.pipeline()
        # FIXME: add guard
        p.hset(NODE + ':' + path, None, None, {'mtime': time.time()})
        p.hset(DATA + ':' + path, None, None, {key    : val})

        # delete old invalid index entry
        if key in self.data:
            # we keep the key index entry around
            p.srem(VALS + ':' + str(self.data[key]), path)
        else:
            # new key: add new key index entry
            p.sadd(KEYS + ':' + str(key), path)

        # always add new value index entry
        p.sadd(VALS + ':' + str(val), path)

        # FIXME: eval return types / values
        _ = p.execute()

        # issue notification about key creation/update
        self.logger.debug("PUB ATTRIBUTE %s [%s=%s]" % (path, key, val))
        self.r.publish(MON,   "ATTRIBUTE %s [%s=%s]" % (path, key, val))

        # update cache
        self.data[key] = val
        self.cache.set(DATA + ':' + path, self.data)

    # --------------------------------------------------------------------------
    #
    def manage_callback(self, key, id, cb, obj):
        # FIXME: this needs a shared lock with the monitor thread!

        self.logger.debug("redis_ns_entry.manage__callback %s: %s" %
                          (self.path, key))

        if self.path not in self.callbacks:
            self.callbacks[self.path] = {}
            if key not in self.callbacks[self.path]:
                self.callbacks[self.path][key] = {}

        if id is None:
            del self.callbacks[self.path][key]

        if cb:
            self.callbacks[self.path][key][id] = [cb, obj]
        else:
            # cb == None: remove that callback
            # FIXME: if we have no callbacks for key, remove [key]
            # FIXME: if we have no keys for path, remove [path]
            # FIXME: if we remove a [path], unsubscribe for its notifications
            del self.callbacks[self.path][key][id]


# ------------------------------------------------------------------------------







__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os

from ...exceptions import *
from ..            import base
from ..cpi         import SYNC_CALL, ASYNC_CALL
from ..cpi         import context as cpi
from ...           import context as api
from ...exceptions import *


######################################################################
#
# adaptor meta data
#
_ADAPTOR_NAME          = 'radical.saga.adaptor.x509'
_ADAPTOR_SCHEMAS       = ['X509']
_ADAPTOR_OPTIONS       = []

# FIXME: complete attribute list
_ADAPTOR_CAPABILITIES  = {
    'attributes'       : [api.TYPE,
                          api.USER_PROXY,
                          api.LIFE_TIME]
}

_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'cfg_options'      : _ADAPTOR_OPTIONS,
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : """This adaptor points to a X509 proxy, or certificate,
                            be used for backend connections.  Note that this
                            context can be created by a MyProxy context instance.""",
    'schemas'          : {'x509' : 'x509 token information.'},
}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.1',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [{
        'type'         : 'radical.saga.Context',
        'class'        : 'ContextX509'
        }
    ]
}


###############################################################################
# The adaptor class

class Adaptor (base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    def __init__ (self) :

        base.Base.__init__ (self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        # there are no default myproxy contexts
        self._default_contexts = []
        self._have_defaults    = False


    def sanity_check (self) :
        pass


    def _get_default_contexts (self) :

        if not self._have_defaults :

            p = os.environ.get ('X509_USER_PROXY', "/tmp/x509up_u%d"  %  os.getuid())

            if  os.path.exists (p) and \
                os.path.isfile (p)     :

                try :
                    fh = open (p)

                except Exception as e:
                    self._logger.warning ("invalid X509 context at %s"  %  p)
                    pass

                else :
                    fh.close ()

                    c = api.Context ('X509')
                    c.user_proxy = p

                    self._logger.info ("default X509 context for proxy at %s"  %  p)

                    self._default_contexts.append (c)
                    self._have_defaults = True

            else:
                if 'X509_USER_PROXY' in os.environ:
                    self._logger.warning ("no X509 context at %s"  %  p)

        # have defaults, and can return them...
        return self._default_contexts



######################################################################
#
# job adaptor class
#
class ContextX509 (cpi.Context) :

    def __init__ (self, api, adaptor) :

        _cpi_base = super  (ContextX509, self)
        _cpi_base.__init__ (api, adaptor)


    @SYNC_CALL
    def init_instance (self, adaptor_state, type) :

        if  not type.lower () in (schema.lower() for schema in _ADAPTOR_SCHEMAS) :
            raise BadParameter("the x509 context only handles x509 contexts!")

        self.get_api ().type = type

        return self.get_api ()


    @SYNC_CALL
    def _initialize (self, session) :

        # make sure we have can access the proxy
        api = self.get_api ()

        if  not api.user_proxy :
            api.user_proxy = "x509up_u%d"  %  os.getuid()

        if  not os.path.exists (api.user_proxy) or \
            not os.path.isfile (api.user_proxy)    :
            raise BadParameter ("X509 proxy does not exist: %s" % api.user_proxy)

        try :
            fh = open (api.user_proxy)
        except Exception as e:
            raise PermissionDenied ("X509 proxy '%s' not readable: %s" \
                                 % (api.user_proxy, str(e))) from e
        else :
            fh.close ()



